package limax.zdb;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import limax.zdb.TRecord.State;

final class LogRecord<K, V> {
	private final static Map<Class<?>, Listenable> cache = new ConcurrentHashMap<>();
	private final Map<K, LogR<K, V>> changed = new HashMap<K, LogR<K, V>>();
	private final Listenable seed;
	private int hasListener = 0;

	LogRecord(Object o) {
		Listenable l;
		if (o == null) {
			l = Listenable.defaultListenable;
		} else if ((l = cache.get(o.getClass())) == null)
			cache.put(o.getClass(), l = Listenable.create(o));
		this.seed = l;
	}

	private LogR<K, V> getLogR(TTable<K, V> table, TRecord<K, V> r) {
		if (hasListener == -1)
			return null;
		Transaction.current().recordLogNotifyTTable(table);
		if (hasListener == 0) {
			if (table.hasListener())
				hasListener = 1;
			else {
				hasListener = -1;
				return null;
			}
		}
		LogR<K, V> lr = changed.get(r.getKey());
		if (null == lr)
			changed.put(r.getKey(), lr = new LogR<K, V>(r.getState(), seed.copy()));
		return lr.update(r.getValue());
	}

	void onChanged(TTable<K, V> table, TRecord<K, V> r, boolean cc, State ss) {
		LogR<K, V> lr = getLogR(table, r);
		if (lr != null && lr.ss == null) {
			lr.cc = cc;
			lr.ss = ss;
		}
	}

	void onChanged(TTable<K, V> table, TRecord<K, V> r, LogNotify ln) {
		LogR<K, V> lr = getLogR(table, r);
		if (lr != null) {
			ln.pop();
			lr.lb.setChanged(ln);
		}
	}

	void logNotify(ListenerMap listenerMap) {
		for (Map.Entry<K, LogR<K, V>> entry : changed.entrySet()) {
			LogR<K, V> lr = entry.getValue();
			lr.lb.logNotify(entry.getKey(), lr.value, lr.getRecordState(), listenerMap);
		}
		hasListener = 0;
		changed.clear();
	}

	private static class LogR<K, V> {
		private final State rs;
		private final Listenable lb;
		private V value;
		private boolean cc;
		private State ss;

		LogR(State rs, Listenable value) {
			this.rs = rs;
			this.lb = value;
		}

		LogR<K, V> update(V value) {
			this.value = value;
			return this;
		}

		private RecordState getRecordState() {
			if (null == ss)
				return RecordState.CHANGED;
			if (cc && State.ADD == ss && State.REMOVE == rs) {
				return RecordState.NONE;
			} else if (!cc && State.ADD == ss && State.REMOVE == rs) {
				return RecordState.REMOVED;
			} else if (State.ADD == ss && State.ADD == rs) {
				return RecordState.ADDED;
			} else if (State.INDB_GET == ss && State.INDB_REMOVE == rs) {
				return RecordState.REMOVED;
			} else if (State.INDB_GET == ss && State.INDB_ADD == rs) {
				return RecordState.ADDED;
			} else if (State.INDB_REMOVE == ss && State.INDB_ADD == rs) {
				return RecordState.ADDED;
			} else if (cc && State.INDB_REMOVE == ss && State.INDB_REMOVE == rs) {
				return RecordState.REMOVED;
			} else if (!cc && State.INDB_REMOVE == ss && State.INDB_REMOVE == rs) {
				return RecordState.NONE;
			} else if (cc && State.REMOVE == ss && State.REMOVE == rs) {
				return RecordState.NONE;
			} else if (cc && State.REMOVE == ss && State.ADD == rs) {
				return RecordState.ADDED;
			} else if (!cc && State.INDB_ADD == ss && State.INDB_ADD == rs) {
				return RecordState.ADDED;
			} else if (!cc && State.INDB_ADD == ss && State.INDB_REMOVE == rs) {
				return RecordState.REMOVED;
			}
			throw new IllegalStateException(
					"AddRemoveState Error! isCreateCache = " + cc + ", SavedState = " + ss + ", State = " + rs);
		}
	}
}
